using namespace std;

#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <iostream>

#include "bitset.h"

#define SEGFAULT() abort()

#define BITSET_BITS \
	( CHAR_BIT * sizeof(size_t) )

#define BITSET_MASK(pos) \
	( ((size_t)1) << ((pos) % BITSET_BITS) )

#define BITSET_WORD(set, pos) \
	( (set)->bits[(pos) / BITSET_BITS] )

#define BITSET_USED(nbits) \
	( ((nbits) + (BITSET_BITS - 1)) / BITSET_BITS )

bitset *bitset_init(size_t nbits) {

  bitset *set;
  set = (bitset*) malloc(sizeof(bitset));
  assert(set);

  set->bits = (size_t*) malloc(BITSET_USED(nbits)*sizeof(size_t));
  set->nbits = nbits;
  
  assert(set->bits);
  
  return set;
}

void bitset_reset(bitset *set) {
  memset(set->bits, (size_t)0, BITSET_USED(set->nbits) * sizeof(*set->bits));
}

//to improve
void bitset_set(bitset *set) {
  for (size_t i = 0; i < set->nbits; i++)
    bitset_set_bit(set,i);
}

//to improve
size_t bitset_count_set(bitset* set) {
  int count = 0;
  
  for (size_t i = 0; i < set->nbits; i++)
    if (bitset_test_bit(set,i))
      count++;

  return count;
}



void bitset_copy(bitset *from, bitset *to) {
  //assumes both from and to have the same bitset size.
  memcpy(to->bits, from->bits, BITSET_USED(to->nbits) * sizeof(*to->bits));
}


void bitset_free(bitset *set) {
  if (set) {
    free(set->bits);
    free(set);
  }
}

void bitset_clear_bit(bitset *set, size_t pos) {
	if (pos >= set->nbits) {
	    SEGFAULT();
	}

	BITSET_WORD(set, pos) &= ~BITSET_MASK(pos);
}

void bitset_set_bit(bitset *set, size_t pos) {
	if (pos >= set->nbits) {
	    SEGFAULT();
	}

	BITSET_WORD(set, pos) |= BITSET_MASK(pos);
}

int bitset_test_bit(bitset *set, size_t pos) {
	if (pos >= set->nbits) {
	    SEGFAULT();
	}

	return (BITSET_WORD(set, pos) & BITSET_MASK(pos)) != 0;
}

void bitset_print(bitset *set) {
  if (!set)
    return;

  cout << "[";
  if (set)
    for (int i = 0; i < set->nbits; i++) 
      if (bitset_test_bit(set, i))
	cout << i << ",";
  cout << "]" << endl;
}

//add set2 to set1
void bitset_union(bitset *set1, bitset *set2)
{
  if (set2 == NULL)
    return;

  for(size_t i = 0; i < BITSET_USED(set1->nbits); i++)
    set1->bits[i] |= set2->bits[i];

  //  for(size_t i = 0; i < set1->nbits; i++)
  //   if (bitset_test_bit(set2,i))
  //   bitset_set_bit(set1,i);
}

//to improve
bitset *bitset_complement(bitset *ref_set, bitset *to_complement)
{
  if (!ref_set || !to_complement)
    return NULL;

  bitset *res = bitset_init(ref_set->nbits);
  bitset_reset(res);

  for (size_t i = 0; i < ref_set->nbits; i++)
    if (bitset_test_bit(ref_set,i) && !bitset_test_bit(to_complement,i))
      bitset_set_bit(res,i);

  return res;
}

//assumes both sets have the same size
//to improve
void bitset_subtract(bitset *from_set, bitset *what)
{
  for (size_t i = 0; i < from_set->nbits; i++)
    if (bitset_test_bit(what,i))
      bitset_clear_bit(from_set,i);
}

void bitset_negate(bitset *ref_set, bitset *to_negate)
{
  for (size_t i = 0; i < ref_set->nbits; i++)
    if (bitset_test_bit(ref_set,i))
      if (bitset_test_bit(to_negate,i))
	bitset_clear_bit(to_negate,i);
      else
	bitset_set_bit(to_negate,i);
}

/* return -1 if set is empty, or the pos of the first set bit otherwise */
int bitset_first_bit_set(bitset *set)
{
  if (!set) return -1;

  int i = 0;
  while (i < set->nbits && !bitset_test_bit(set, i))
    i++;
  if (i < set->nbits)
    return i;
  return -1;
}


/* return -1 if set is not singleton, otherise the pos of the set bit */
int bitset_test_singleton(bitset *set)
{
  if (!set) return -1;

  int pos = -1;
  int i = 0;
  int how_many = 0;
  while (i < set->nbits && how_many < 2)
    {
      if(bitset_test_bit(set, i)) {
	how_many++;
	pos = i;
      }
    i++;
    }
  if (how_many == 1)
    return pos;
  return -1;
}

int bitset_test_empty(bitset *set)
{
  if (!set) return 1;


  size_t i = 0;
  while (i < BITSET_USED(set->nbits) && set->bits[i] == (size_t)0)
    i++;

  return (i == BITSET_USED(set->nbits));

  /*
  int i = 0;
  while (i < set->nbits && !bitset_test_bit(set, i))
    i++;
  return (i == set->nbits);
  */
}


